#!/usr/bin/env python3

from netCDF4 import Dataset
import matplotlib
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import cartopy.feature as cfeature
import numpy as np
from grid_utils import get_grid


def plot_weights(ax, weightsfile, src_grid, tgt_grid,
                 src_idx=None, tgt_idx=None, quiver_kwargs={}):
    weights = Dataset(weightsfile, "r")
    if weights.version != "yac weight file 1.0":
        print("WARNING: You are using an incompatible weight file version\n" +
              weights.version + " != yac weight file 1.0")
    src_points = np.empty([2, 0])
    num_src_fields = weights.num_src_fields if "num_src_fields" in weights.variables else 1
    for src_field in range(num_src_fields):
        locstr = bytes(np.array(weights["src_locations"][src_field, :])).decode().rstrip("\0")
        if locstr == "CELL":
            src_points = np.hstack([src_points, np.stack([src_grid.clon,
                                                          src_grid.clat])])
        elif locstr == "CORNER":
            src_points = np.hstack([src_points, np.stack([src_grid.vlon,
                                                          src_grid.vlat])])
        elif locstr == "EDGE":
            src_points = np.hstack([src_points, np.stack([src_grid.vlon,
                                                          src_grid.vlat])])
        else:
            raise f"Unknown location string {locstr}"
    locstr = bytes(np.array(weights["dst_location"])).decode().rstrip("\0")
    if locstr == "CELL":
        tgt_points = np.stack([tgt_grid.clon, tgt_grid.clat])
    elif locstr == "CORNER":
        tgt_points = np.stack([tgt_grid.vlon, tgt_grid.vlat])
    elif locstr == "EDGE":
        tgt_points = np.stack([tgt_grid.elon, tgt_grid.elat])
    else:
        raise f"Unknown location string {locstr}"

    yac_weights_format_address_offset = 1
    src_adr = np.asarray(weights["src_address"])-yac_weights_format_address_offset-src_grid.idx_offset
    tgt_adr = np.asarray(weights["dst_address"])-yac_weights_format_address_offset-tgt_grid.idx_offset
    extent = ax.get_extent(crs=ccrs.PlateCarree())
    mask = ((src_points[0, src_adr] >= extent[0]) * (src_points[0, src_adr] <= extent[1]) *
            (src_points[1, src_adr] >= extent[2]) * (src_points[1, src_adr] <= extent[3])) + \
           ((tgt_points[0, tgt_adr] >= extent[0]) * (tgt_points[0, tgt_adr] <= extent[1]) *
            (tgt_points[1, tgt_adr] >= extent[2]) * (tgt_points[1, tgt_adr] <= extent[3]))
    if src_idx is not None:
        mask *= (src_adr == (src_idx-src_grid.idx_offset))
    if tgt_idx is not None:
        mask *= (tgt_adr == (tgt_idx-tgt_grid.idx_offset))
    src_adr = src_adr[mask]
    tgt_adr = tgt_adr[mask]

    ax_proj = ax.projection
    src_t = ax_proj.transform_points(ccrs.PlateCarree(),
                                     src_points[0, src_adr], src_points[1, src_adr])
    tgt_t = ax_proj.transform_points(ccrs.PlateCarree(),
                                     tgt_points[0, tgt_adr], tgt_points[1, tgt_adr])
    uv = tgt_t-src_t

    c = weights["remap_matrix"][mask, 0]

    norm = matplotlib.colors.Normalize()
    cm = matplotlib.cm.Oranges  # decide for colormap
    ax.quiver(src_t[:, 0], src_t[:, 1],
              uv[:, 0], uv[:, 1], angles='xy', scale_units='xy', scale=1,
              color=cm(norm(c)),
              **quiver_kwargs)
    sm = matplotlib.cm.ScalarMappable(cmap=cm, norm=norm)
    sm.set_array([])
    plt.colorbar(sm, ax=ax)

    # add label if restricted to one cell:
    if src_idx is not None:
        ax.set_title(f"Interpolation for source index {src_idx}\nSum of weights: {sum(c):.8f}")
        for x, y, t in zip(tgt_points[0, tgt_adr], tgt_points[1, tgt_adr], c):
            ax.text(x, y, f"{t:.3}", transform=ccrs.PlateCarree())
    if tgt_idx is not None:
        ax.set_title(f"Interpolation for target index {tgt_idx}\nSum of weights: {sum(c):.8f}")
        for x, y, t in zip(src_points[0, src_adr], src_points[1, src_adr], c):
            ax.text(x, y, f"{t:.3}", transform=ccrs.PlateCarree())


def cell_extent(grid, idx, zoom=4):
    print(f"cell {idx}: ", grid.clon[idx-grid.idx_offset],
          grid.clat[idx-grid.idx_offset])
    vidx = grid.vertex_of_cell[:, idx-grid.idx_offset]
    e = np.array([np.min(grid.vlon[vidx]), np.max(grid.vlon[vidx]),
                  np.min(grid.vlat[vidx]), np.max(grid.vlat[vidx])])
    c = np.array([0.5*(e[0]+e[1]), 0.5*(e[0]+e[1]),
                  0.5*(e[2]+e[3]), 0.5*(e[2]+e[3])])
    return (e-c)*zoom + c


def main(source_grid, target_grid, weights_file, center=None,
         source_idx=None, target_idx=None, zoom=2,
         label_src_grid=None, label_tgt_grid=None):
    src_grid = get_grid(source_grid)
    if target_grid is not None:
        tgt_grid = get_grid(target_grid)
    fig = plt.figure(figsize=[10, 10])
    if source_idx is not None:
        center = [src_grid.clon[source_idx-src_grid.idx_offset],
                  src_grid.clat[source_idx-src_grid.idx_offset]]
    elif target_idx is not None:
        center = [tgt_grid.clon[target_idx-tgt_grid.idx_offset],
                  tgt_grid.clat[target_idx-tgt_grid.idx_offset]]

    proj = ccrs.Orthographic(*center)
    ax = fig.add_subplot(1, 1, 1, projection=proj)

    if source_idx is not None:
        extent = cell_extent(src_grid, source_idx, zoom)
        ax.set_extent(extent, crs=ccrs.PlateCarree())
    elif target_idx is not None:
        extent = cell_extent(tgt_grid, target_idx, zoom)
        ax.set_extent(extent, crs=ccrs.PlateCarree())
    else:
        ax.set_extent([-1000000*zoom, 1000000*zoom, -1000000*zoom, 1000000*zoom], crs=proj)

    # Put a background image on for nice sea rendering.
#    ax.stock_img()
    ax.set_facecolor("#9ddbff")
    feature = cfeature.NaturalEarthFeature(name='land', category='physical',
                                           scale='10m',
                                           edgecolor='#000000', facecolor='#cbe7be')
    ax.add_feature(feature, zorder=-999)

    src_grid.plot(ax, label=label_src_grid, plot_kwargs={"color": "green"})
    if target_grid is not None:
        tgt_grid.plot(ax, label=label_tgt_grid, plot_kwargs={"color": "blue"})

    if target_grid is not None and weights_file is not None:
        plot_weights(ax, weights_file, src_grid, tgt_grid,
                     source_idx, target_idx, quiver_kwargs={})
    plt.show()


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(prog="plot_weights.py",
                                     description="""
                                     plot grids and yac weights file.

                                     Grids can be specified either a filename.
                                     This is iterpreted like an ICON grid file or
                                     a string like "g360,180" from which a structured
                                     grid is generated. Use a capital G for 1-based
                                     indexing and a small g for 0-basid indexing
                                     followed by the resolution (lon,lat). Optionally
                                     you can add an extent by adding further 4 numbers
                                     (min_lon, max_lon, min_lat, max_lat). E.g.
                                     g100,100,-50,-45,-5,5
                                     """)
    parser.add_argument("source_grid", type=str,
                        help="source grid (netCDF file or [gG]lon,lat[,min_lon,max_lon,min_lat,max_lat])")
    parser.add_argument("target_grid", type=str, nargs='?',
                        default=None,
                        help="target grid (netCDF file or [gG]lon,lat[,min_lon,max_lon,min_lat,max_lat])")
    parser.add_argument("weights_file", type=str, help="YAC weights file", nargs='?',
                        default=None)
    parser.add_argument("--center", "-c", type=float, nargs=2, help="center of the orthografic projection",
                        default=(0, 0), metavar=("LON", "LON"))
    parser.add_argument("--source_idx", "-s", type=int,
                        help="index of source cell to focus")
    parser.add_argument("--target_idx", "-t", type=int,
                        help="index of target cell to focus")
    parser.add_argument("--zoom", "-z", type=float, default=2,
                        help="zoom around the cell")
    parser.add_argument("--label_src_grid", type=str, default=None,
                        choices=("vertex", "edge", "cell"),
                        help="Add labels at the source grid")
    parser.add_argument("--label_tgt_grid", type=str, default=None,
                        choices=("vertex", "edge", "cell"),
                        help="Add labels at the source grid")
    args = parser.parse_args()
    main(**args.__dict__)
